import json
from time import sleep
from unittest import SkipTest
from unittest.mock import patch
from uuid import uuid4

import boto3
import pytest
from botocore.client import ClientError

from moto import mock_aws, settings
from moto.core import DEFAULT_ACCOUNT_ID
from moto.s3 import s3_backends
from moto.s3.models import FakeBucket
from moto.s3.responses import DEFAULT_REGION_NAME
from tests.test_s3 import empty_bucket, s3_aws_verified


@mock_aws
def test_put_bucket_logging():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    wrong_region_client = boto3.client("s3", region_name="us-west-2")
    bucket_name = "mybucket"
    log_bucket = "logbucket"
    wrong_region_bucket = "wrongregionlogbucket"
    s3_client.create_bucket(Bucket=bucket_name)
    # Adding the ACL for log-delivery later...
    s3_client.create_bucket(Bucket=log_bucket)
    wrong_region_client.create_bucket(
        Bucket=wrong_region_bucket,
        CreateBucketConfiguration={"LocationConstraint": "us-west-2"},
    )

    # No logging config:
    result = s3_client.get_bucket_logging(Bucket=bucket_name)
    assert not result.get("LoggingEnabled")

    # A log-bucket that doesn't exist:
    with pytest.raises(ClientError) as err:
        s3_client.put_bucket_logging(
            Bucket=bucket_name,
            BucketLoggingStatus={
                "LoggingEnabled": {"TargetBucket": "IAMNOTREAL", "TargetPrefix": ""}
            },
        )
    assert err.value.response["Error"]["Code"] == "InvalidTargetBucketForLogging"

    # A log-bucket that's missing the proper ACLs for LogDelivery:
    with pytest.raises(ClientError) as err:
        s3_client.put_bucket_logging(
            Bucket=bucket_name,
            BucketLoggingStatus={
                "LoggingEnabled": {"TargetBucket": log_bucket, "TargetPrefix": ""}
            },
        )
    assert err.value.response["Error"]["Code"] == "InvalidTargetBucketForLogging"
    assert "log-delivery" in err.value.response["Error"]["Message"]

    # Add the proper "log-delivery" ACL to the log buckets:
    bucket_owner = s3_client.get_bucket_acl(Bucket=log_bucket)["Owner"]
    for bucket in [log_bucket, wrong_region_bucket]:
        s3_client.put_bucket_acl(
            Bucket=bucket,
            AccessControlPolicy={
                "Grants": [
                    {
                        "Grantee": {
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                            "Type": "Group",
                        },
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                            "Type": "Group",
                        },
                        "Permission": "READ_ACP",
                    },
                    {
                        "Grantee": {"Type": "CanonicalUser", "ID": bucket_owner["ID"]},
                        "Permission": "FULL_CONTROL",
                    },
                ],
                "Owner": bucket_owner,
            },
        )

    # A log-bucket that's in the wrong region:
    with pytest.raises(ClientError) as err:
        s3_client.put_bucket_logging(
            Bucket=bucket_name,
            BucketLoggingStatus={
                "LoggingEnabled": {
                    "TargetBucket": wrong_region_bucket,
                    "TargetPrefix": "",
                }
            },
        )
    assert err.value.response["Error"]["Code"] == "CrossLocationLoggingProhibitted"

    # Correct logging:
    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {
                "TargetBucket": log_bucket,
                "TargetPrefix": f"{bucket_name}/",
            }
        },
    )
    result = s3_client.get_bucket_logging(Bucket=bucket_name)
    assert result["LoggingEnabled"]["TargetBucket"] == log_bucket
    assert result["LoggingEnabled"]["TargetPrefix"] == f"{bucket_name}/"
    assert not result["LoggingEnabled"].get("TargetGrants")

    # And disabling:
    s3_client.put_bucket_logging(Bucket=bucket_name, BucketLoggingStatus={})
    assert not s3_client.get_bucket_logging(Bucket=bucket_name).get("LoggingEnabled")

    # And enabling with multiple target grants:
    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {
                "TargetBucket": log_bucket,
                "TargetPrefix": f"{bucket_name}/",
                "TargetGrants": [
                    {
                        "Grantee": {
                            "ID": "SOMEIDSTRINGHERE9238748923734823917498237489237409123840983274",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "READ",
                    },
                    {
                        "Grantee": {
                            "ID": "SOMEIDSTRINGHERE9238748923734823917498237489237409123840983274",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "WRITE",
                    },
                ],
            }
        },
    )

    result = s3_client.get_bucket_logging(Bucket=bucket_name)
    assert len(result["LoggingEnabled"]["TargetGrants"]) == 2
    assert (
        result["LoggingEnabled"]["TargetGrants"][0]["Grantee"]["ID"]
        == "SOMEIDSTRINGHERE9238748923734823917498237489237409123840983274"
    )

    # Test with just 1 grant:
    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {
                "TargetBucket": log_bucket,
                "TargetPrefix": f"{bucket_name}/",
                "TargetGrants": [
                    {
                        "Grantee": {
                            "ID": "SOMEIDSTRINGHERE9238748923734823917498237489237409123840983274",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "READ",
                    }
                ],
            }
        },
    )
    result = s3_client.get_bucket_logging(Bucket=bucket_name)
    assert len(result["LoggingEnabled"]["TargetGrants"]) == 1

    # With an invalid grant:
    with pytest.raises(ClientError) as err:
        s3_client.put_bucket_logging(
            Bucket=bucket_name,
            BucketLoggingStatus={
                "LoggingEnabled": {
                    "TargetBucket": log_bucket,
                    "TargetPrefix": f"{bucket_name}/",
                    "TargetGrants": [
                        {
                            "Grantee": {
                                "ID": (
                                    "SOMEIDSTRINGHERE9238748923734823917498"
                                    "237489237409123840983274"
                                ),
                                "Type": "CanonicalUser",
                            },
                            "Permission": "NOTAREALPERM",
                        }
                    ],
                }
            },
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"


@mock_aws
def test_log_file_is_created():
    # Create necessary buckets
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "mybucket"
    log_bucket = "logbucket"
    s3_client.create_bucket(Bucket=bucket_name)
    s3_client.create_bucket(Bucket=log_bucket)

    # Enable logging
    bucket_owner = s3_client.get_bucket_acl(Bucket=log_bucket)["Owner"]
    s3_client.put_bucket_acl(
        Bucket=log_bucket,
        AccessControlPolicy={
            "Grants": [
                {
                    "Grantee": {
                        "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        "Type": "Group",
                    },
                    "Permission": "WRITE",
                },
                {
                    "Grantee": {
                        "URI": "http://acs.amazonaws.com/groups/s3/LogDelivery",
                        "Type": "Group",
                    },
                    "Permission": "READ_ACP",
                },
                {
                    "Grantee": {"Type": "CanonicalUser", "ID": bucket_owner["ID"]},
                    "Permission": "FULL_CONTROL",
                },
            ],
            "Owner": bucket_owner,
        },
    )
    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {
                "TargetBucket": log_bucket,
                "TargetPrefix": f"{bucket_name}/",
            }
        },
    )

    # Make some requests against the source bucket
    s3_client.put_object(Bucket=bucket_name, Key="key1", Body=b"")
    s3_client.put_object(Bucket=bucket_name, Key="key2", Body=b"data")

    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {"TargetBucket": log_bucket, "TargetPrefix": ""}
        },
    )
    s3_client.list_objects_v2(Bucket=bucket_name)

    # Verify files are created in the target (logging) bucket
    keys = [k["Key"] for k in s3_client.list_objects_v2(Bucket=log_bucket)["Contents"]]
    assert len([k for k in keys if k.startswith("mybucket/")]) == 3
    assert len([k for k in keys if not k.startswith("mybucket/")]) == 1

    # Verify (roughly) files have the correct content
    contents = [
        s3_client.get_object(Bucket=log_bucket, Key=key)["Body"].read().decode("utf-8")
        for key in keys
    ]
    assert any(c for c in contents if bucket_name in c)
    assert any(c for c in contents if "REST.GET.BUCKET" in c)
    assert any(c for c in contents if "REST.PUT.BUCKET" in c)


@mock_aws
def test_invalid_bucket_logging_when_permissions_are_false():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "mybucket"
    log_bucket = "logbucket"
    s3_client.create_bucket(Bucket=bucket_name)
    s3_client.create_bucket(Bucket=log_bucket)
    with patch(
        "moto.s3.models.FakeBucket._log_permissions_enabled_policy", return_value=False
    ), patch(
        "moto.s3.models.FakeBucket._log_permissions_enabled_acl", return_value=False
    ):
        with pytest.raises(ClientError) as err:
            s3_client.put_bucket_logging(
                Bucket=bucket_name,
                BucketLoggingStatus={
                    "LoggingEnabled": {"TargetBucket": log_bucket, "TargetPrefix": ""}
                },
            )
        assert err.value.response["Error"]["Code"] == "InvalidTargetBucketForLogging"
        assert "log-delivery" in err.value.response["Error"]["Message"]


@mock_aws
def test_valid_bucket_logging_when_permissions_are_true():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "mybucket"
    log_bucket = "logbucket"
    s3_client.create_bucket(Bucket=bucket_name)
    s3_client.create_bucket(Bucket=log_bucket)
    with patch(
        "moto.s3.models.FakeBucket._log_permissions_enabled_policy", return_value=True
    ), patch(
        "moto.s3.models.FakeBucket._log_permissions_enabled_acl", return_value=True
    ):
        s3_client.put_bucket_logging(
            Bucket=bucket_name,
            BucketLoggingStatus={
                "LoggingEnabled": {
                    "TargetBucket": log_bucket,
                    "TargetPrefix": f"{bucket_name}/",
                }
            },
        )
        result = s3_client.get_bucket_logging(Bucket=bucket_name)
        assert result["LoggingEnabled"]["TargetBucket"] == log_bucket
        assert result["LoggingEnabled"]["TargetPrefix"] == f"{bucket_name}/"


@mock_aws
def test_bucket_policy_not_set():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_backend = s3_backends[DEFAULT_ACCOUNT_ID]["global"]

    log_bucket = "log_bucket"
    s3_client.create_bucket(Bucket=log_bucket)
    log_bucket_obj = s3_backend.get_bucket(log_bucket)

    assert (
        FakeBucket._log_permissions_enabled_policy(
            target_bucket=log_bucket_obj, target_prefix=""
        )
        is False
    )


@mock_aws
def test_bucket_policy_principal():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_backend = s3_backends[DEFAULT_ACCOUNT_ID]["global"]

    log_bucket = "log_bucket"
    s3_client.create_bucket(Bucket=log_bucket)
    log_bucket_obj = s3_backend.get_bucket(log_bucket)

    invalid_principal_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "not_logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(invalid_principal_policy)
    )
    assert (
        FakeBucket._log_permissions_enabled_policy(
            target_bucket=log_bucket_obj, target_prefix=""
        )
        is False
    )

    s3_client.delete_bucket_policy(Bucket=log_bucket)
    valid_principal_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(valid_principal_policy)
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix=""
    )


@mock_aws
def test_bucket_policy_effect():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_backend = s3_backends[DEFAULT_ACCOUNT_ID]["global"]

    log_bucket = "log_bucket"
    s3_client.create_bucket(Bucket=log_bucket)
    log_bucket_obj = s3_backend.get_bucket(log_bucket)
    deny_effect_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Deny",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(deny_effect_policy)
    )
    assert (
        FakeBucket._log_permissions_enabled_policy(
            target_bucket=log_bucket_obj, target_prefix=""
        )
        is False
    )

    s3_client.delete_bucket_policy(Bucket=log_bucket)
    allow_effect_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(allow_effect_policy)
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix=""
    )


@mock_aws
def test_bucket_policy_action():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_backend = s3_backends[DEFAULT_ACCOUNT_ID]["global"]

    log_bucket = "log_bucket"
    s3_client.create_bucket(Bucket=log_bucket)
    log_bucket_obj = s3_backend.get_bucket(log_bucket)
    non_put_object_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:GetObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(non_put_object_policy)
    )
    assert (
        FakeBucket._log_permissions_enabled_policy(
            target_bucket=log_bucket_obj, target_prefix=""
        )
        is False
    )

    s3_client.delete_bucket_policy(Bucket=log_bucket)
    put_object_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(Bucket=log_bucket, Policy=json.dumps(put_object_policy))
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix=""
    )


@mock_aws
def test_bucket_policy_resource():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Can't patch permission logic in ServerMode")

    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    s3_backend = s3_backends[DEFAULT_ACCOUNT_ID]["global"]

    log_bucket = "log_bucket"
    s3_client.create_bucket(Bucket=log_bucket)
    log_bucket_obj = s3_backend.get_bucket(log_bucket)
    entire_bucket_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(entire_bucket_policy)
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix=""
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix="prefix"
    )

    s3_client.delete_bucket_policy(Bucket=log_bucket)
    bucket_level_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(bucket_level_policy)
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix=""
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix="prefix"
    )

    s3_client.delete_bucket_policy(Bucket=log_bucket)
    specfic_prefix_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket}/prefix*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket, Policy=json.dumps(specfic_prefix_policy)
    )
    assert (
        FakeBucket._log_permissions_enabled_policy(
            target_bucket=log_bucket_obj, target_prefix=""
        )
        is False
    )
    assert FakeBucket._log_permissions_enabled_policy(
        target_bucket=log_bucket_obj, target_prefix="prefix"
    )


@s3_aws_verified
@pytest.mark.aws_verified
def test_put_logging_w_bucket_policy_no_prefix(bucket_name=None):
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    log_bucket_name = f"{uuid4()}"
    s3_client.create_bucket(Bucket=log_bucket_name)
    bucket_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket_name}/*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket_name, Policy=json.dumps(bucket_policy)
    )
    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {"TargetBucket": log_bucket_name, "TargetPrefix": ""}
        },
    )
    result = s3_client.get_bucket_logging(Bucket=bucket_name)
    # Logging Config is not immediately available
    for _ in range(5):
        if "LoggingEnabled" not in result:
            sleep(1)
            result = s3_client.get_bucket_logging(Bucket=bucket_name)
    assert result["LoggingEnabled"]["TargetBucket"] == log_bucket_name
    assert result["LoggingEnabled"]["TargetPrefix"] == ""

    empty_bucket(s3_client, log_bucket_name)
    s3_client.delete_bucket(Bucket=log_bucket_name)


@s3_aws_verified
@pytest.mark.aws_verified
def test_put_logging_w_bucket_policy_w_prefix(bucket_name=None):
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

    log_bucket_name = f"{uuid4()}"
    s3_client.create_bucket(Bucket=log_bucket_name)
    prefix = "some-prefix"
    bucket_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "S3ServerAccessLogsPolicy",
                "Effect": "Allow",
                "Principal": {"Service": "logging.s3.amazonaws.com"},
                "Action": ["s3:PutObject"],
                "Resource": f"arn:aws:s3:::{log_bucket_name}/{prefix}*",
            }
        ],
    }
    s3_client.put_bucket_policy(
        Bucket=log_bucket_name, Policy=json.dumps(bucket_policy)
    )
    s3_client.put_bucket_logging(
        Bucket=bucket_name,
        BucketLoggingStatus={
            "LoggingEnabled": {"TargetBucket": log_bucket_name, "TargetPrefix": prefix}
        },
    )
    result = s3_client.get_bucket_logging(Bucket=bucket_name)
    # Logging Config is not immediately available
    for _ in range(5):
        if "LoggingEnabled" not in result:
            sleep(1)
            result = s3_client.get_bucket_logging(Bucket=bucket_name)
    assert result["LoggingEnabled"]["TargetBucket"] == log_bucket_name
    assert result["LoggingEnabled"]["TargetPrefix"] == prefix

    empty_bucket(s3_client, log_bucket_name)
    s3_client.delete_bucket(Bucket=log_bucket_name)



import datetime
from unittest import SkipTest
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws, settings
from moto.core.utils import iso_8601_datetime_with_milliseconds
from tests import EXAMPLE_AMI_ID


@mock_aws
def test_request_spot_instances():
    conn = boto3.client("ec2", "us-east-1")
    vpc = conn.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]
    subnet = conn.create_subnet(
        VpcId=vpc["VpcId"], CidrBlock="10.0.0.0/16", AvailabilityZone="us-east-1a"
    )["Subnet"]
    subnet_id = subnet["SubnetId"]

    sec_name_1 = str(uuid4())
    sec_name_2 = str(uuid4())
    conn.create_security_group(GroupName=sec_name_1, Description="description")
    conn.create_security_group(GroupName=sec_name_2, Description="description")

    start_dt = datetime.datetime(2013, 1, 1).replace(tzinfo=datetime.timezone.utc)
    end_dt = datetime.datetime(2013, 1, 2).replace(tzinfo=datetime.timezone.utc)
    start = iso_8601_datetime_with_milliseconds(start_dt)
    end = iso_8601_datetime_with_milliseconds(end_dt)

    with pytest.raises(ClientError) as ex:
        request = conn.request_spot_instances(
            SpotPrice="0.5",
            InstanceCount=1,
            Type="one-time",
            ValidFrom=start,
            ValidUntil=end,
            LaunchGroup="the-group",
            AvailabilityZoneGroup="my-group",
            LaunchSpecification={
                "ImageId": EXAMPLE_AMI_ID,
                "KeyName": "test",
                "SecurityGroups": [sec_name_1, sec_name_2],
                "UserData": "some test data",
                "InstanceType": "m1.small",
                "Placement": {"AvailabilityZone": "us-east-1c"},
                "KernelId": "test-kernel",
                "RamdiskId": "test-ramdisk",
                "Monitoring": {"Enabled": True},
                "SubnetId": subnet_id,
            },
            DryRun=True,
        )
    assert ex.value.response["Error"]["Code"] == "DryRunOperation"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 412
    assert (
        ex.value.response["Error"]["Message"]
        == "An error occurred (DryRunOperation) when calling the RequestSpotInstances operation: Request would have succeeded, but DryRun flag is set"
    )

    request = conn.request_spot_instances(
        SpotPrice="0.5",
        InstanceCount=1,
        Type="one-time",
        ValidFrom=start,
        ValidUntil=end,
        LaunchGroup="the-group",
        AvailabilityZoneGroup="my-group",
        LaunchSpecification={
            "ImageId": EXAMPLE_AMI_ID,
            "KeyName": "test",
            "SecurityGroups": [sec_name_1, sec_name_2],
            "UserData": "some test data",
            "InstanceType": "m1.small",
            "Placement": {"AvailabilityZone": "us-east-1c"},
            "KernelId": "test-kernel",
            "RamdiskId": "test-ramdisk",
            "Monitoring": {"Enabled": True},
            "SubnetId": subnet_id,
        },
    )
    request_id = request["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    all_requests = conn.describe_spot_instance_requests()["SpotInstanceRequests"]
    requests = [r for r in all_requests if r["SpotInstanceRequestId"] == request_id]
    assert len(requests) == 1
    request = requests[0]

    assert request["State"] == "active"
    assert request["SpotPrice"] == "0.500000"
    assert request["Type"] == "one-time"
    assert request["ValidFrom"] == start_dt
    assert request["ValidUntil"] == end_dt
    assert request["LaunchGroup"] == "the-group"
    assert request["AvailabilityZoneGroup"] == "my-group"

    launch_spec = request["LaunchSpecification"]
    security_group_names = [
        group["GroupName"] for group in launch_spec["SecurityGroups"]
    ]
    assert set(security_group_names) == set([sec_name_1, sec_name_2])

    assert launch_spec["ImageId"] == EXAMPLE_AMI_ID
    assert launch_spec["KeyName"] == "test"
    assert launch_spec["InstanceType"] == "m1.small"
    assert launch_spec["KernelId"] == "test-kernel"
    assert launch_spec["RamdiskId"] == "test-ramdisk"
    assert launch_spec["SubnetId"] == subnet_id


@mock_aws
def test_request_spot_instances_default_arguments():
    """
    Test that moto set the correct default arguments
    """
    conn = boto3.client("ec2", "us-east-1")

    request = conn.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    request_id = request["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    all_requests = conn.describe_spot_instance_requests()["SpotInstanceRequests"]
    requests = [r for r in all_requests if r["SpotInstanceRequestId"] == request_id]
    assert len(requests) == 1
    request = requests[0]

    assert request["State"] == "active"
    assert request["SpotPrice"] == "0.500000"
    assert request["Type"] == "one-time"
    assert "ValidFrom" not in request
    assert "ValidUntil" not in request
    assert "LaunchGroup" not in request
    assert "AvailabilityZoneGroup" not in request
    assert request["InstanceInterruptionBehavior"] == "terminate"

    launch_spec = request["LaunchSpecification"]

    security_group_names = [
        group["GroupName"] for group in launch_spec["SecurityGroups"]
    ]
    assert security_group_names == ["default"]

    assert launch_spec["ImageId"] == EXAMPLE_AMI_ID
    assert "KeyName" not in request
    assert launch_spec["InstanceType"] == "m1.small"
    assert "KernelId" not in request
    assert "RamdiskId" not in request
    assert "SubnetId" not in request


@mock_aws
def test_cancel_spot_instance_request():
    client = boto3.client("ec2", region_name="us-west-1")

    rsi = client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    spot_id = rsi["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    requests = client.describe_spot_instance_requests(SpotInstanceRequestIds=[spot_id])[
        "SpotInstanceRequests"
    ]
    assert len(requests) == 1
    request = requests[0]
    assert "CreateTime" in request
    assert request["Type"] == "one-time"
    assert "SpotInstanceRequestId" in request
    assert request["SpotPrice"] == "0.500000"
    assert request["LaunchSpecification"]["ImageId"] == EXAMPLE_AMI_ID

    with pytest.raises(ClientError) as ex:
        client.cancel_spot_instance_requests(
            SpotInstanceRequestIds=[request["SpotInstanceRequestId"]], DryRun=True
        )
    assert ex.value.response["Error"]["Code"] == "DryRunOperation"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 412
    assert (
        ex.value.response["Error"]["Message"]
        == "An error occurred (DryRunOperation) when calling the CancelSpotInstanceRequests operation: Request would have succeeded, but DryRun flag is set"
    )

    client.cancel_spot_instance_requests(
        SpotInstanceRequestIds=[request["SpotInstanceRequestId"]]
    )

    requests = client.describe_spot_instance_requests(SpotInstanceRequestIds=[spot_id])[
        "SpotInstanceRequests"
    ]
    assert len(requests) == 0


@mock_aws
def test_request_spot_instances_fulfilled():
    """
    Test that moto correctly fullfills a spot instance request
    """
    client = boto3.client("ec2", region_name="us-east-1")

    request = client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    request_id = request["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    requests = client.describe_spot_instance_requests(
        SpotInstanceRequestIds=[request_id]
    )["SpotInstanceRequests"]
    assert len(requests) == 1
    request = requests[0]

    assert request["State"] == "active"


@mock_aws
def test_tag_spot_instance_request():
    """
    Test that moto correctly tags a spot instance request
    """
    client = boto3.client("ec2", region_name="us-west-1")

    request = client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    request_id = request["SpotInstanceRequests"][0]["SpotInstanceRequestId"]
    client.create_tags(
        Resources=[request_id],
        Tags=[{"Key": "tag1", "Value": "value1"}, {"Key": "tag2", "Value": "value2"}],
    )

    requests = client.describe_spot_instance_requests(
        SpotInstanceRequestIds=[request_id]
    )["SpotInstanceRequests"]
    assert len(requests) == 1
    request = requests[0]

    assert len(request["Tags"]) == 2
    assert {"Key": "tag1", "Value": "value1"} in request["Tags"]
    assert {"Key": "tag2", "Value": "value2"} in request["Tags"]


@mock_aws
def test_get_all_spot_instance_requests_filtering():
    """
    Test that moto correctly filters spot instance requests
    """
    client = boto3.client("ec2", region_name="us-west-1")

    request1 = client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    request1_id = request1["SpotInstanceRequests"][0]["SpotInstanceRequestId"]
    request2 = client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    request2_id = request2["SpotInstanceRequests"][0]["SpotInstanceRequestId"]
    client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    tag_value1 = str(uuid4())
    client.create_tags(
        Resources=[request1_id],
        Tags=[{"Key": "tag1", "Value": tag_value1}, {"Key": "tag2", "Value": "value2"}],
    )
    client.create_tags(
        Resources=[request2_id],
        Tags=[{"Key": "tag1", "Value": tag_value1}, {"Key": "tag2", "Value": "wrong"}],
    )

    requests = client.describe_spot_instance_requests(
        Filters=[{"Name": "state", "Values": ["failed"]}]
    )["SpotInstanceRequests"]
    r_ids = [r["SpotInstanceRequestId"] for r in requests]
    assert request1_id not in r_ids
    assert request2_id not in r_ids

    requests = client.describe_spot_instance_requests(
        Filters=[{"Name": "state", "Values": ["active"]}]
    )["SpotInstanceRequests"]
    r_ids = [r["SpotInstanceRequestId"] for r in requests]
    assert request1_id in r_ids
    assert request2_id in r_ids

    requests = client.describe_spot_instance_requests(
        Filters=[{"Name": "tag:tag1", "Values": [tag_value1]}]
    )["SpotInstanceRequests"]
    assert len(requests) == 2

    requests = client.describe_spot_instance_requests(
        Filters=[
            {"Name": "tag:tag1", "Values": [tag_value1]},
            {"Name": "tag:tag2", "Values": ["value2"]},
        ]
    )["SpotInstanceRequests"]
    assert len(requests) == 1


@mock_aws
@pytest.mark.filterwarnings("ignore")
def test_request_spot_instances_instance_lifecycle():
    if settings.TEST_SERVER_MODE:
        # Currently no easy way to check which instance was created by request_spot_instance
        # And we can't just pick the first instance in ServerMode and expect it to be the right one
        raise SkipTest("ServerMode is not guaranteed to be empty")
    client = boto3.client("ec2", region_name="us-east-1")
    client.request_spot_instances(SpotPrice="0.5")

    response = client.describe_instances()

    instance = response["Reservations"][0]["Instances"][0]
    assert instance["InstanceLifecycle"] == "spot"


@mock_aws
@pytest.mark.filterwarnings("ignore")
def test_request_spot_instances_with_tags():
    client = boto3.client("ec2", region_name="us-east-1")
    request = client.request_spot_instances(
        SpotPrice="0.5",
        TagSpecifications=[
            {
                "ResourceType": "spot-instances-request",
                "Tags": [{"Key": "k", "Value": "v"}],
            }
        ],
    )

    request_id = request["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    request = client.describe_spot_instance_requests(
        SpotInstanceRequestIds=[request_id]
    )["SpotInstanceRequests"][0]
    assert request["Tags"] == [{"Key": "k", "Value": "v"}]


@mock_aws
def test_launch_spot_instance_instance_lifecycle():
    client = boto3.client("ec2", region_name="us-east-1")

    kwargs = {
        "KeyName": "foobar",
        "ImageId": EXAMPLE_AMI_ID,
        "MinCount": 1,
        "MaxCount": 1,
        "InstanceType": "c4.2xlarge",
        "TagSpecifications": [
            {"ResourceType": "instance", "Tags": [{"Key": "key", "Value": "val"}]},
        ],
        "InstanceMarketOptions": {"MarketType": "spot"},
    }

    instance = client.run_instances(**kwargs)["Instances"][0]
    instance_id = instance["InstanceId"]

    response = client.describe_instances(InstanceIds=[instance_id])
    instance = response["Reservations"][0]["Instances"][0]
    assert instance["InstanceLifecycle"] == "spot"


@mock_aws
def test_launch_instance_instance_lifecycle():
    client = boto3.client("ec2", region_name="us-east-1")

    kwargs = {
        "KeyName": "foobar",
        "ImageId": EXAMPLE_AMI_ID,
        "MinCount": 1,
        "MaxCount": 1,
        "InstanceType": "c4.2xlarge",
        "TagSpecifications": [
            {"ResourceType": "instance", "Tags": [{"Key": "key", "Value": "val"}]},
        ],
    }

    instance = client.run_instances(**kwargs)["Instances"][0]
    instance_id = instance["InstanceId"]

    response = client.describe_instances(InstanceIds=[instance_id])
    instance = response["Reservations"][0]["Instances"][0]
    assert instance.get("InstanceLifecycle") is None


@mock_aws
def test_spot_price_history():
    client = boto3.client("ec2", region_name="us-east-1")
    # test filter
    response = client.describe_spot_price_history(
        Filters=[
            {"Name": "availability-zone", "Values": ["us-east-1a"]},
            {"Name": "instance-type", "Values": ["t3a.micro"]},
        ]
    )
    price = response["SpotPriceHistory"][0]
    assert price["InstanceType"] == "t3a.micro"
    assert price["AvailabilityZone"] == "us-east-1a"

    # test instance types
    i_types = ["t3a.micro", "t3.micro"]
    response = client.describe_spot_price_history(InstanceTypes=i_types)
    price = response["SpotPriceHistory"][0]
    assert price["InstanceType"] in i_types


@mock_aws
def test_request_spot_instances__instance_should_exist():
    client = boto3.client("ec2", region_name="us-east-1")
    request = client.request_spot_instances(
        SpotPrice="0.5", LaunchSpecification={"ImageId": EXAMPLE_AMI_ID}
    )
    request_id = request["SpotInstanceRequests"][0]["SpotInstanceRequestId"]

    request = client.describe_spot_instance_requests(
        SpotInstanceRequestIds=[request_id]
    )["SpotInstanceRequests"][0]
    assert "InstanceId" in request
    instance_id = request["InstanceId"]

    response = client.describe_instances(InstanceIds=[instance_id])
    instance = response["Reservations"][0]["Instances"][0]
    assert instance["InstanceId"] == instance_id
    assert instance["ImageId"] == EXAMPLE_AMI_ID



"""
Ensure that the responses module plays nice with our mocks
"""

from unittest import SkipTest, TestCase

import boto3
import requests
import responses

from moto import mock_aws, settings
from moto.core.models import override_responses_real_send
from moto.core.versions import RESPONSES_VERSION
from moto.utilities.distutils_version import LooseVersion


class TestResponsesModule(TestCase):
    def setUp(self) -> None:
        if settings.TEST_SERVER_MODE:
            raise SkipTest("No point in testing responses-decorator in ServerMode")

    @mock_aws
    @responses.activate
    def test_moto_first(self) -> None:  # type: ignore[misc]
        """
        Verify we can activate a user-defined `responses` on top of our Moto mocks
        """
        self.moto_responses_compatibility()

    @responses.activate
    @mock_aws
    def test_moto_second(self) -> None:
        """
        Verify we can load Moto after activating a `responses`-mock
        """
        self.moto_responses_compatibility()

    def moto_responses_compatibility(self) -> None:
        responses.add(
            responses.GET, url="http://127.0.0.1/lkdsfjlkdsa", json={"a": "4"}
        )
        s3 = boto3.client("s3", region_name="us-east-1")
        s3.create_bucket(Bucket="mybucket")
        s3.put_object(Bucket="mybucket", Key="name", Body="value")
        s3.get_object(Bucket="mybucket", Key="name")["Body"].read()
        with requests.get("http://127.0.0.1/lkdsfjlkdsa") as r:
            assert r.json() == {"a": "4"}

    @responses.activate
    def test_moto_as_late_as_possible(self) -> None:
        """
        Verify we can load moto after registering a response
        """
        responses.add(
            responses.GET, url="http://127.0.0.1/lkdsfjlkdsa", json={"a": "4"}
        )
        with mock_aws():
            s3 = boto3.client("s3", region_name="us-east-1")
            s3.create_bucket(Bucket="mybucket")
            s3.put_object(Bucket="mybucket", Key="name", Body="value")
            # This mock exists within Moto
            with requests.get("http://127.0.0.1/lkdsfjlkdsa") as r:
                assert r.json() == {"a": "4"}

        # And outside of Moto
        with requests.get("http://127.0.0.1/lkdsfjlkdsa") as r:
            assert r.json() == {"a": "4"}


@mock_aws
class TestResponsesMockWithPassThru(TestCase):
    """
    https://github.com/getmoto/moto/issues/6417
    """

    def setUp(self) -> None:
        if RESPONSES_VERSION < LooseVersion("0.24.0"):
            raise SkipTest("Can only test this with responses >= 0.24.0")

        self.r_mock = responses.RequestsMock(assert_all_requests_are_fired=True)
        override_responses_real_send(self.r_mock)
        self.r_mock.start()
        self.r_mock.add_passthru("http://ip.jsontest.com")

    def tearDown(self) -> None:
        self.r_mock.stop()
        self.r_mock.reset()
        override_responses_real_send(None)

    def http_requests(self) -> str:
        # Mock this website
        requests.post("https://example.org")

        # Passthrough this website
        assert requests.get("http://ip.jsontest.com").status_code == 200

        return "OK"

    def aws_and_http_requests(self) -> str:
        ddb = boto3.client("dynamodb", "us-east-1")
        assert ddb.list_tables()["TableNames"] == []
        self.http_requests()
        return "OK"

    def test_http_requests(self) -> None:
        self.r_mock.add(responses.POST, "https://example.org", status=200)
        self.assertEqual("OK", self.http_requests())

    def test_aws_and_http_requests(self) -> None:
        self.r_mock.add(responses.POST, "https://example.org", status=200)
        self.assertEqual("OK", self.aws_and_http_requests())



import json

import moto.server as server


def test_es_list():
    backend = server.create_backend_app("es")
    test_client = backend.test_client()

    resp = test_client.get("/2015-01-01/domain")
    assert resp.status_code == 200
    assert json.loads(resp.data) == {"DomainNames": []}



import json

from jsonpath_ng.ext import parse
from jsonpath_ng.jsonpath import Index

from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class JSONPathUtils:
    @staticmethod
    def extract_json(path: str, data: json) -> json:
        input_expr = parse(path)
        matches = input_expr.find(data)
        if not matches:
            raise RuntimeError(
                f"The JSONPath {path} could not be found in the input {to_json_str(data)}"
            )
        if (
            len(matches) > 1
            or isinstance(matches[0].path, Index)
            or (matches[0].context and isinstance(matches[0].context.path, Index))
        ):
            value = [match.value for match in matches]
        else:
            value = matches[0].value
        return value



import json
from typing import Dict, List, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.moto_api._internal import mock_random as random


class SageMakerRuntimeBackend(BaseBackend):
    """Implementation of SageMakerRuntime APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.async_results: Dict[str, Dict[str, Tuple[str, str]]] = {}
        self.results: Dict[str, Dict[bytes, Tuple[str, str, str, str]]] = {}
        self.results_queue: List[Tuple[str, str, str, str]] = []
        self.async_results_queue: List[Tuple[bool, str]] = []

    def invoke_endpoint(
        self, endpoint_name: str, unique_repr: bytes
    ) -> Tuple[str, str, str, str]:
        """
        This call will return static data by default.

        You can use a dedicated API to override this, by configuring a queue of expected results.

        A request to `get_query_results` will take the first result from that queue. Subsequent requests using the same details will return the same result. Other requests using a different QueryExecutionId will take the next result from the queue, or return static data if the queue is empty.

        Configuring this queue by making an HTTP request to `/moto-api/static/sagemaker/endpoint-results`. An example invocation looks like this:

        .. sourcecode:: python

            expected_results = {
                "account_id": "123456789012",  # This is the default - can be omitted
                "region": "us-east-1",  # This is the default - can be omitted
                "results": [
                    {
                         "Body": "first body",
                         "ContentType": "text/xml",
                         "InvokedProductionVariant": "prod",
                         "CustomAttributes": "my_attr",
                     },
                    # other results as required
                ],
            }
            requests.post(
                "http://motoapi.amazonaws.com/moto-api/static/sagemaker/endpoint-results",
                json=expected_results,
            )

            client = boto3.client("sagemaker", region_name="us-east-1")
            details = client.invoke_endpoint(EndpointName="asdf", Body="qwer")

        """
        if endpoint_name not in self.results:
            self.results[endpoint_name] = {}
        if unique_repr in self.results[endpoint_name]:
            return self.results[endpoint_name][unique_repr]
        if self.results_queue:
            self.results[endpoint_name][unique_repr] = self.results_queue.pop(0)
        else:
            self.results[endpoint_name][unique_repr] = (
                "body",
                "content_type",
                "invoked_production_variant",
                "custom_attributes",
            )
        return self.results[endpoint_name][unique_repr]

    def invoke_endpoint_async(
        self, endpoint_name: str, input_location: str
    ) -> Tuple[str, str]:
        """
        This call will return static data by default.

        You can use a dedicated API to override this, by configuring a queue of expected results.
        Subsequent requests using the same `InputLocation` will return the same result.
        Other requests using a different `InputLocation` will take the next result from the queue.
        Please note that for backward compatibility, if the async queue is empty, the sync queue will be used, if configured.

        Configuring this queue by making an HTTP request to `/moto-api/static/sagemaker/async-endpoint-results`.
        An example invocation looks like this:

        .. sourcecode:: python

            expected_results = {
                "account_id": "123456789012",  # This is the default - can be omitted
                "region": "us-east-1",  # This is the default - can be omitted
                "results": [
                    {
                        "data": json.dumps({"first": "output"}),
                    },
                    {
                        "is_failure": True,
                        "data": "second inference failed",
                    },
                    # other results as required
                ],
            }
            requests.post(
                "http://motoapi.amazonaws.com/moto-api/static/sagemaker/async-endpoint-results",
                json=expected_results,
            )

            client = boto3.client("sagemaker-runtime", region_name="us-east-1")
            details = client.invoke_endpoint_async(EndpointName="asdf", InputLocation="qwer")

        """
        if endpoint_name not in self.async_results:
            self.async_results[endpoint_name] = {}
        if input_location in self.async_results[endpoint_name]:
            return self.async_results[endpoint_name][input_location]
        if self.async_results_queue:
            is_failure, data = self.async_results_queue.pop(0)
        elif self.results_queue:  # Backward compatibility
            is_failure = False
            body, _type, variant, attrs = self.results_queue.pop(0)
            data = json.dumps(
                {
                    "Body": body,
                    "ContentType": _type,
                    "InvokedProductionVariant": variant,
                    "CustomAttributes": attrs,
                }
            )
        else:
            is_failure = False
            data = json.dumps({"default": "response"})

        output_bucket = f"sagemaker-output-{random.uuid4()}"
        output_location = "response.json"
        failure_location = "failure.json"
        result_location = failure_location if is_failure else output_location
        from moto.s3.models import s3_backends

        s3_backend = s3_backends[self.account_id][self.partition]
        s3_backend.create_bucket(output_bucket, region_name=self.region_name)
        s3_backend.put_object(
            output_bucket, result_location, value=data.encode("utf-8")
        )

        self.async_results[endpoint_name][input_location] = (
            f"s3://{output_bucket}/{output_location}",
            f"s3://{output_bucket}/{failure_location}",
        )
        return self.async_results[endpoint_name][input_location]


sagemakerruntime_backends = BackendDict(SageMakerRuntimeBackend, "sagemaker-runtime")



from typing import Final

from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class CauseDecl(EvalComponent):
    value: Final[str]

    def __init__(self, value: str):
        self.value = value

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(self.value)



from decimal import Decimal

import boto3
import pytest
from botocore.exceptions import ClientError

from . import dynamodb_aws_verified


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_update_different_map_elements_in_single_request(table_name=None):
    # https://github.com/getmoto/moto/issues/5552
    dynamodb = boto3.resource("dynamodb", region_name="us-east-1")

    record = {
        "pk": "example_id",
        "d": {"hello": "h", "world": "w"},
    }
    table = dynamodb.Table(table_name)
    table.put_item(Item=record)
    updated = table.update_item(
        Key={"pk": "example_id"},
        UpdateExpression="set d.hello = :h, d.world = :w",
        ExpressionAttributeValues={":h": "H", ":w": "W"},
        ReturnValues="ALL_NEW",
    )
    assert updated["Attributes"] == {
        "pk": "example_id",
        "d": {"hello": "H", "world": "W"},
    }

    # Use UpdateExpression that contains a new-line
    # https://github.com/getmoto/moto/issues/7127
    table.update_item(
        Key={"pk": "example_id"},
        UpdateExpression=(
            """
            ADD 
              MyTotalCount :MyCount
            """
        ),
        ExpressionAttributeValues={":MyCount": 5},
    )
    assert table.get_item(Key={"pk": "example_id"})["Item"]["MyTotalCount"] == 5


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_update_item_add_float(table_name=None):
    table = boto3.resource("dynamodb", "us-east-1").Table(table_name)

    # DECIMAL - DECIMAL
    table.put_item(Item={"pk": "foo", "amount": Decimal(100), "nr": 5})
    table.update_item(
        Key={"pk": "foo"},
        UpdateExpression="ADD amount :delta",
        ExpressionAttributeValues={":delta": -Decimal("88.3")},
    )
    assert table.scan()["Items"][0]["amount"] == Decimal("11.7")

    # DECIMAL + DECIMAL
    table.update_item(
        Key={"pk": "foo"},
        UpdateExpression="ADD amount :delta",
        ExpressionAttributeValues={":delta": Decimal("25.41")},
    )
    assert table.scan()["Items"][0]["amount"] == Decimal("37.11")

    # DECIMAL + INT
    table.update_item(
        Key={"pk": "foo"},
        UpdateExpression="ADD amount :delta",
        ExpressionAttributeValues={":delta": 6},
    )
    assert table.scan()["Items"][0]["amount"] == Decimal("43.11")

    # INT + INT
    table.update_item(
        Key={"pk": "foo"},
        UpdateExpression="ADD nr :delta",
        ExpressionAttributeValues={":delta": 1},
    )
    assert table.scan()["Items"][0]["nr"] == Decimal("6")

    # INT + DECIMAL
    table.update_item(
        Key={"pk": "foo"},
        UpdateExpression="ADD nr :delta",
        ExpressionAttributeValues={":delta": Decimal("25.41")},
    )
    assert table.scan()["Items"][0]["nr"] == Decimal("31.41")


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_delete_non_existing_item(table_name=None):
    table = boto3.resource("dynamodb", "us-east-1").Table(table_name)

    name = "name"
    user_name = "karl"

    # Initial item does not contain users
    initial_item = {"pk": name}
    table.put_item(Item=initial_item)

    # We can remove a (non-existing) user without it failing
    table.update_item(
        Key={"pk": name},
        UpdateExpression="DELETE #users :users",
        ExpressionAttributeValues={":users": {user_name}},
        ExpressionAttributeNames={"#users": "users"},
        ReturnValues="ALL_NEW",
    )
    assert table.get_item(Key={"pk": name})["Item"] == {"pk": "name"}

    # IF the item does exist
    table.update_item(
        Key={"pk": name},
        UpdateExpression="ADD #users :delta",
        ExpressionAttributeNames={"#users": "users"},
        ExpressionAttributeValues={":delta": {user_name}},
    )
    assert table.get_item(Key={"pk": name})["Item"] == {"pk": "name", "users": {"karl"}}

    # We can delete a non-existing item from it
    table.update_item(
        Key={"pk": name},
        UpdateExpression="DELETE #users :users",
        ExpressionAttributeValues={":users": {f"{user_name}2"}},
        ExpressionAttributeNames={"#users": "users"},
        ReturnValues="ALL_NEW",
    )
    assert table.get_item(Key={"pk": name})["Item"] == {"pk": "name", "users": {"karl"}}

    table.update_item(
        Key={"pk": name},
        UpdateExpression="DELETE #users :users",
        ExpressionAttributeValues={":users": {user_name}},
        ExpressionAttributeNames={"#users": "users"},
        ReturnValues="ALL_NEW",
    )
    assert table.get_item(Key={"pk": name})["Item"] == {"pk": "name"}


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_update_item_add_empty_set(table_name=None):
    dynamodb = boto3.client("dynamodb", "us-east-1")

    # Add to non-existing attribute
    dynamodb.put_item(TableName=table_name, Item={"pk": {"S": "foo"}})
    with pytest.raises(ClientError) as exc:
        dynamodb.update_item(
            TableName=table_name,
            Key={"pk": {"S": "foo"}},
            UpdateExpression="ADD stringset :emptySet",
            ExpressionAttributeValues={":emptySet": {"SS": ()}},
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        "ExpressionAttributeValues contains invalid value: One or more parameter values were invalid: An string set  may not be empty"
        in err["Message"]
    )

    assert dynamodb.scan(TableName=table_name)["Items"] == [{"pk": {"S": "foo"}}]

    # Still not allowed when the attribute exists
    dynamodb.put_item(
        TableName=table_name, Item={"pk": {"S": "foo"}, "stringset": {"SS": ("item1",)}}
    )
    with pytest.raises(ClientError) as exc:
        dynamodb.update_item(
            TableName=table_name,
            Key={"pk": {"S": "foo"}},
            UpdateExpression="ADD stringset :emptySet",
            ExpressionAttributeValues={":emptySet": {"SS": ()}},
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        "ExpressionAttributeValues contains invalid value: One or more parameter values were invalid: An string set  may not be empty"
        in err["Message"]
    )

    assert dynamodb.scan(TableName=table_name)["Items"] == [
        {"pk": {"S": "foo"}, "stringset": {"SS": ["item1"]}}
    ]


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_update_item_with_empty_values(table_name=None):
    dynamodb = boto3.client("dynamodb", "us-east-1")

    dynamodb.put_item(TableName=table_name, Item={"pk": {"S": "foo"}})
    with pytest.raises(ClientError) as exc:
        dynamodb.update_item(
            TableName=table_name,
            Key={"pk": {"S": "foo"}},
            UpdateExpression="SET #d = :s",
            ExpressionAttributeNames={"#d": "d"},
            ExpressionAttributeValues={},
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert err["Message"] == "ExpressionAttributeValues must not be empty"


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_update_item_with_empty_expression(table_name=None):
    dynamodb = boto3.client("dynamodb", "us-east-1")

    dynamodb.put_item(TableName=table_name, Item={"pk": {"S": "foo"}})
    with pytest.raises(ClientError) as exc:
        dynamodb.update_item(
            TableName=table_name,
            Key={"pk": {"S": "foo"}},
            UpdateExpression="",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"] == "Invalid UpdateExpression: The expression can not be empty;"
    )



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_db_cluster():
    client = boto3.client("neptune", region_name="us-east-2")
    resp = client.create_db_cluster(DBClusterIdentifier="cluster-id", Engine="neptune")[
        "DBCluster"
    ]
    assert resp["DBClusterIdentifier"] == "cluster-id"
    assert "DbClusterResourceId" in resp
    assert "DBClusterArn" in resp
    assert resp["Engine"] == "neptune"
    assert "EngineVersion" in resp
    assert resp["StorageEncrypted"] is False
    assert resp["DBClusterParameterGroup"].startswith("default.neptune")
    assert "Endpoint" in resp
    assert "cluster-" in resp["DbClusterResourceId"]
    assert resp["AvailabilityZones"] == ["us-east-2a", "us-east-2b", "us-east-2c"]
    assert "ServerlessV2ScalingConfiguration" not in resp

    # Double check this cluster is not available in another region
    europe_client = boto3.client("neptune", region_name="eu-west-2")
    assert len(europe_client.describe_db_clusters()["DBClusters"]) == 0


@mock_aws
def test_create_db_cluster__with_additional_params():
    client = boto3.client("neptune", region_name="us-east-1")
    resp = client.create_db_cluster(
        DBClusterIdentifier="cluster-id",
        Engine="neptune",
        EngineVersion="1.1.0.1",
        StorageEncrypted=False,
        DBClusterParameterGroupName="myprm",
        KmsKeyId="key",
        ServerlessV2ScalingConfiguration={"MinCapacity": 1.0, "MaxCapacity": 2.0},
        DatabaseName="sth",
    )["DBCluster"]
    assert resp["StorageEncrypted"] is False
    assert resp["DBClusterParameterGroup"] == "myprm"
    assert resp["EngineVersion"] == "1.1.0.1"
    assert resp["KmsKeyId"] == "key"
    assert resp["ServerlessV2ScalingConfiguration"] == {
        "MinCapacity": 1.0,
        "MaxCapacity": 2.0,
    }
    assert resp["DatabaseName"] == "sth"


@mock_aws
def test_describe_db_clusters():
    client = boto3.client("neptune", region_name="ap-southeast-1")
    assert client.describe_db_clusters()["DBClusters"] == []

    client.create_db_cluster(DBClusterIdentifier="cluster-id", Engine="neptune")

    clusters = client.describe_db_clusters(DBClusterIdentifier="cluster-id")[
        "DBClusters"
    ]
    assert len(clusters) == 1
    assert clusters[0]["DBClusterIdentifier"] == "cluster-id"
    assert clusters[0]["Engine"] == "neptune"


@mock_aws
def test_delete_db_cluster():
    client = boto3.client("neptune", region_name="ap-southeast-1")

    client.create_db_cluster(DBClusterIdentifier="cluster-id", Engine="neptune")
    client.delete_db_cluster(DBClusterIdentifier="cluster-id")

    assert client.describe_db_clusters()["DBClusters"] == []


@mock_aws
def test_delete_unknown_db_cluster():
    client = boto3.client("neptune", region_name="ap-southeast-1")

    with pytest.raises(ClientError) as exc:
        client.delete_db_cluster(DBClusterIdentifier="unknown-id")
    err = exc.value.response["Error"]
    assert err["Code"] == "DBClusterNotFoundFault"


@mock_aws
def test_modify_db_cluster():
    client = boto3.client("neptune", region_name="us-east-1")
    client.create_db_cluster(DBClusterIdentifier="cluster-id", Engine="neptune")
    resp = client.modify_db_cluster(
        DBClusterIdentifier="cluster-id",
        EngineVersion="1.1.0.1",
        DBClusterParameterGroupName="myprm",
        PreferredBackupWindow="window",
    )["DBCluster"]
    assert resp["DBClusterParameterGroup"] == "myprm"
    assert resp["EngineVersion"] == "1.1.0.1"
    assert resp["PreferredBackupWindow"] == "window"


@mock_aws
def test_start_db_cluster():
    client = boto3.client("neptune", region_name="us-east-2")
    client.create_db_cluster(DBClusterIdentifier="cluster-id", Engine="neptune")
    client.stop_db_cluster(DBClusterIdentifier="cluster-id")
    cluster = client.start_db_cluster(DBClusterIdentifier="cluster-id")["DBCluster"]
    assert cluster["Status"] == "started"



from datetime import datetime, timedelta, timezone
from time import sleep
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core.utils import utcnow
from tests.test_cloudwatch import cloudwatch_aws_verified


@mock_aws
def test_get_metric_data__no_metric_data_or_expression():
    utc_now = datetime.now(tz=timezone.utc)
    cloudwatch = boto3.client("cloudwatch", "eu-west-1")
    with pytest.raises(ClientError) as exc:
        cloudwatch.get_metric_data(
            MetricDataQueries=[{"Id": "result1", "Label": "e1"}],
            StartTime=utc_now - timedelta(minutes=5),
            EndTime=utc_now,
            ScanBy="TimestampDescending",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationError"
    assert (
        err["Message"]
        == "The parameter MetricDataQueries.member.1.MetricStat is required.\n"
    )


@mock_aws
def test_get_metric_data_with_simple_expression():
    utc_now = datetime.now(tz=timezone.utc)
    cloudwatch = boto3.client("cloudwatch", "eu-west-1")
    namespace = "my_namespace/"
    cloudwatch.put_metric_data(
        Namespace=namespace,
        MetricData=[
            {
                "MetricName": "metric",
                "Value": 25,
                "Unit": "Bytes",
            },
        ],
    )
    # get_metric_data 1
    results = cloudwatch.get_metric_data(
        MetricDataQueries=[
            {
                "Id": "result1",
                "Expression": "totalBytes",
                "Label": "e1",
            },
            {
                "Id": "totalBytes",
                "MetricStat": {
                    "Metric": {"Namespace": namespace, "MetricName": "metric"},
                    "Period": 60,
                    "Stat": "Sum",
                    "Unit": "Bytes",
                },
                "ReturnData": False,
            },
        ],
        StartTime=utc_now - timedelta(minutes=5),
        EndTime=utc_now + timedelta(minutes=5),
        ScanBy="TimestampDescending",
    )["MetricDataResults"]
    #
    assert len(results) == 1
    assert results[0]["Id"] == "result1"
    assert results[0]["Label"] == "e1"
    assert results[0]["Values"] == [25.0]


@cloudwatch_aws_verified
@pytest.mark.aws_verified
def test_get_metric_data_with_expressive_expression():
    cloudwatch = boto3.client("cloudwatch", "eu-west-1")
    my_namespace = f"Test/FakeNamespace_{str(uuid4())[0:6]}"
    cloudwatch.put_metric_data(
        Namespace=my_namespace,
        MetricData=[
            {
                "MetricName": "ApiResponse",
                "Dimensions": [
                    {"Name": "Path", "Value": "/my/path"},
                    {"Name": "Audience", "Value": "my audience"},
                ],
                "Timestamp": utcnow(),
                "StatisticValues": {
                    "Maximum": 10,
                    "Minimum": 0,
                    "Sum": 10,
                    "SampleCount": 10,
                },
            }
        ],
    )

    expression = f"SELECT SUM(ApiResponse) FROM \"{my_namespace}\" WHERE Path = '/my/path' GROUP BY Audience"
    start_time = utcnow() - timedelta(hours=1)
    end_time = utcnow() + timedelta(hours=1)

    for _ in range(60):
        result = cloudwatch.get_metric_data(
            StartTime=start_time,
            EndTime=end_time,
            MetricDataQueries=[
                {
                    "Id": "query",
                    "Period": 60,
                    "Expression": expression,
                }
            ],
        )["MetricDataResults"]
        if result:
            assert len(result) == 1
            assert result[0]["Id"] == "query"
            # If the Label is not supplied, AWS uses:
            #     the group value as the query name, if the query contains a GROUP BY clause
            #     the Id if no GROUP BY clause exists
            # But Moto doesn't understand expressions, so we always use the Id
            assert result[0]["Label"]
            assert result[0]["Values"] == [10.0]
            assert result[0]["StatusCode"] == "Complete"
            return
        else:
            sleep(5)
    assert False, "Should have found metrics within 5 minutes"



import boto3
import pytest

from moto import mock_aws


@pytest.fixture(scope="function")
def account_id():
    """Return the current Account ID. Will reach out to AWS if so configured."""
    from . import allow_aws_request

    if allow_aws_request():
        identity = boto3.client("sts", "us-east-1").get_caller_identity()
        yield identity["Account"]
    else:
        with mock_aws():
            identity = boto3.client("sts", "us-east-1").get_caller_identity()
            yield identity["Account"]



from datetime import timedelta

import boto3
import cryptography
import pytest
from botocore.exceptions import ClientError
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509 import Name, NameAttribute
from cryptography.x509.oid import NameOID

from moto.core.utils import utcnow
from tests.test_iam import iam_aws_verified


@iam_aws_verified(create_user=True)
@pytest.mark.aws_verified
def test_signing_certs(user_name=None):
    client = boto3.client("iam", region_name="us-east-1")
    certificate = create_certificate()

    # Upload the cert:
    resp = client.upload_signing_certificate(
        UserName=user_name, CertificateBody=certificate
    )["Certificate"]
    cert_id = resp["CertificateId"]

    assert resp["UserName"] == user_name
    assert resp["Status"] == "Active"
    assert resp["CertificateBody"] == certificate
    assert resp["CertificateId"]

    # Update:
    client.update_signing_certificate(
        UserName=user_name, CertificateId=cert_id, Status="Inactive"
    )

    # List the certs:
    resp = client.list_signing_certificates(UserName=user_name)["Certificates"]
    assert len(resp) == 1
    assert resp[0]["CertificateBody"] == certificate
    assert resp[0]["Status"] == "Inactive"  # Changed with the update call above.

    # Delete:
    client.delete_signing_certificate(UserName=user_name, CertificateId=cert_id)


@iam_aws_verified(create_user=True)
@pytest.mark.aws_verified
def test_create_too_many_certificates(user_name=None):
    client = boto3.client("iam", region_name="us-east-1")
    certificate1 = create_certificate()
    certificate2 = create_certificate()
    certificate3 = create_certificate()

    # Upload two certs
    cert_id1 = client.upload_signing_certificate(
        UserName=user_name, CertificateBody=certificate1
    )["Certificate"]["CertificateId"]
    cert_id2 = client.upload_signing_certificate(
        UserName=user_name, CertificateBody=certificate2
    )["Certificate"]["CertificateId"]
    assert cert_id1 != cert_id2

    # Verify that a third certificate exceeds the limit
    with pytest.raises(ClientError) as exc:
        client.upload_signing_certificate(
            UserName=user_name, CertificateBody=certificate3
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "LimitExceeded"
    assert err["Message"] == "Cannot exceed quota for CertificatesPerUser: 2"


@iam_aws_verified(create_user=True)
def test_retrieve_cert_details_using_credentials_report(user_name=None):
    """
    AWS caches the Credentials Report for 4 hours
    Once you generate the report, you can request the report over and over again, but you'll always get that same report back in the next 4 hours
    # That makes it slightly impossible to verify this against AWS - that's why there is no `aws_verified`-marker
    """
    client = boto3.client("iam", region_name="us-east-1")
    certificate1 = create_certificate()
    certificate2 = create_certificate()

    # Upload the cert:
    cert_id1 = client.upload_signing_certificate(
        UserName=user_name, CertificateBody=certificate1
    )["Certificate"]["CertificateId"]

    result = client.generate_credential_report()
    while result["State"] != "COMPLETE":
        result = client.generate_credential_report()
    report = client.get_credential_report()["Content"].decode("utf-8")

    our_line = next(line for line in report.split("\n") if line.startswith(user_name))
    cert1_active = our_line.split(",")[-4]
    assert cert1_active == "true"
    cert2_active = our_line.split(",")[-2]
    assert cert2_active == "false"

    client.upload_signing_certificate(UserName=user_name, CertificateBody=certificate2)

    result = client.generate_credential_report()
    while result["State"] != "COMPLETE":
        result = client.generate_credential_report()
    report = client.get_credential_report()["Content"].decode("utf-8")
    our_line = next(line for line in report.split("\n") if line.startswith(user_name))

    cert1_active = our_line.split(",")[-4]
    assert cert1_active == "true"
    cert2_active = our_line.split(",")[-2]
    assert cert2_active == "true"

    # Set Certificate to Inactive
    client.update_signing_certificate(
        UserName=user_name, CertificateId=cert_id1, Status="Inactive"
    )

    # Verify the credential report is updated
    result = client.generate_credential_report()
    while result["State"] != "COMPLETE":
        result = client.generate_credential_report()
    report = client.get_credential_report()["Content"].decode("utf-8")
    our_line = next(line for line in report.split("\n") if line.startswith(user_name))

    cert1_active = our_line.split(",")[-4]
    assert cert1_active == "false"
    cert2_active = our_line.split(",")[-2]
    assert cert2_active == "true"


@iam_aws_verified()
@pytest.mark.aws_verified
def test_upload_cert_for_unknown_user(user_name=None):
    client = boto3.client("iam", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.upload_signing_certificate(
            UserName="notauser", CertificateBody=create_certificate()
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert err["Message"] == "The user with name notauser cannot be found."

    with pytest.raises(ClientError) as exc:
        client.update_signing_certificate(
            UserName="notauser",
            CertificateId="asdfasdfasdfasdfasdfasdfasdasdf",
            Status="Inactive",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert err["Message"] == "The user with name notauser cannot be found."

    with pytest.raises(ClientError) as exc:
        client.list_signing_certificates(UserName="notauser")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert err["Message"] == "The user with name notauser cannot be found."

    with pytest.raises(ClientError):
        client.delete_signing_certificate(UserName="notauser", CertificateId="x" * 32)
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert err["Message"] == "The user with name notauser cannot be found."


@iam_aws_verified(create_user=True)
@pytest.mark.aws_verified
def test_upload_invalid_certificate(user_name=None):
    client = boto3.client("iam", region_name="us-east-1")
    with pytest.raises(ClientError) as ce:
        client.upload_signing_certificate(
            UserName=user_name, CertificateBody="notacert"
        )
    assert ce.value.response["Error"]["Code"] == "MalformedCertificate"


@iam_aws_verified(create_user=True)
@pytest.mark.aws_verified
def test_update_unknown_certificate(user_name=None):
    client = boto3.client("iam", region_name="us-east-1")
    fake_id_name = "x" * 32
    with pytest.raises(ClientError) as ce:
        client.update_signing_certificate(
            UserName=user_name, CertificateId=fake_id_name, Status="Inactive"
        )
    err = ce.value.response["Error"]

    assert err["Message"] == f"The Certificate with id {fake_id_name} cannot be found."


def create_certificate():
    key = rsa.generate_private_key(public_exponent=65537, key_size=2028)
    cert_subject = [NameAttribute(NameOID.COMMON_NAME, "iam.amazonaws.com")]
    issuer = [
        NameAttribute(NameOID.COUNTRY_NAME, "US"),
        NameAttribute(NameOID.ORGANIZATION_NAME, "Amazon"),
        NameAttribute(NameOID.COMMON_NAME, "Amazon RSA 2048 M01"),
    ]
    cert = (
        cryptography.x509.CertificateBuilder()
        .subject_name(Name(cert_subject))
        .issuer_name(Name(issuer))
        .public_key(key.public_key())
        .serial_number(cryptography.x509.random_serial_number())
        .not_valid_before(utcnow())
        .not_valid_after(utcnow() + timedelta(days=365))
        .sign(key, hashes.SHA256())
    )
    return cert.public_bytes(serialization.Encoding.PEM).decode("utf-8")



from moto.core.exceptions import JsonRESTError


class DoesNotExistException(JsonRESTError):
    code = 400

    def __init__(self, window_id: str):
        super().__init__(
            "DoesNotExistException", f"Maintenance window {window_id} does not exist"
        )


class InvalidFilterKey(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidFilterKey", message)


class InvalidFilterOption(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidFilterOption", message)


class InvalidFilterValue(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidFilterValue", message)


class InvalidResourceId(JsonRESTError):
    code = 400

    def __init__(self) -> None:
        super().__init__("InvalidResourceId", "Invalid Resource Id")


class InvalidResourceType(JsonRESTError):
    code = 400

    def __init__(self) -> None:
        super().__init__("InvalidResourceType", "Invalid Resource Type")


class ParameterNotFound(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ParameterNotFound", message)


class ParameterVersionNotFound(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ParameterVersionNotFound", message)


class ParameterVersionLabelLimitExceeded(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ParameterVersionLabelLimitExceeded", message)


class ValidationException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class DocumentAlreadyExists(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("DocumentAlreadyExists", message)


class DocumentPermissionLimit(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("DocumentPermissionLimit", message)


class InvalidPermissionType(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidPermissionType", message)


class InvalidDocument(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidDocument", message)


class InvalidDocumentOperation(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidDocumentOperation", message)


class AccessDeniedException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("AccessDeniedException", message)


class InvalidDocumentContent(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidDocumentContent", message)


class InvalidDocumentVersion(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidDocumentVersion", message)


class DuplicateDocumentVersionName(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("DuplicateDocumentVersionName", message)


class DuplicateDocumentContent(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("DuplicateDocumentContent", message)


class ParameterMaxVersionLimitExceeded(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ParameterMaxVersionLimitExceeded", message)


class ParameterAlreadyExists(JsonRESTError):
    code = 400

    def __init__(self) -> None:
        super().__init__(
            "ParameterAlreadyExists",
            "The parameter already exists. To overwrite this value, set the overwrite option in the request to true.",
        )



from typing import Final

from moto.stepfunctions.parser.asl.component.intrinsic.function.function import Function
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.component.state.fail.cause_decl import CauseDecl
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.parse.intrinsic.intrinsic_parser import (
    IntrinsicParser,
)
from moto.stepfunctions.parser.asl.utils.json_path import JSONPathUtils

_STRING_RETURN_FUNCTIONS = {
    typ.name()
    for typ in [
        StatesFunctionNameType.Format,
        StatesFunctionNameType.JsonToString,
        StatesFunctionNameType.ArrayGetItem,
        StatesFunctionNameType.Base64Decode,
        StatesFunctionNameType.Base64Encode,
        StatesFunctionNameType.Hash,
        StatesFunctionNameType.UUID,
    ]
}


class CausePath(CauseDecl): ...


class CausePathJsonPath(CausePath):
    def _eval_body(self, env: Environment) -> None:
        current_output = env.stack[-1]
        cause = JSONPathUtils.extract_json(self.value, current_output)
        env.stack.append(cause)


class CausePathIntrinsicFunction(CausePath):
    function: Final[Function]

    def __init__(self, value: str) -> None:
        super().__init__(value=value)
        self.function = IntrinsicParser.parse(value)
        if self.function.name.name not in _STRING_RETURN_FUNCTIONS:
            raise ValueError(
                f"Unsupported Intrinsic Function for CausePath declaration: '{self.value}'."
            )

    def _eval_body(self, env: Environment) -> None:
        self.function.eval(env=env)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional

from moto.appmesh.dataclasses.route import Route
from moto.appmesh.dataclasses.shared import Metadata, Status


@dataclass
class PortMapping:
    port: Optional[int]
    protocol: Optional[str]


@dataclass
class VirtualRouterSpec:
    listeners: List[Dict[Literal["port_mapping"], PortMapping]]


@dataclass
class VirtualRouter:
    mesh_name: str
    metadata: Metadata
    spec: VirtualRouterSpec
    status: Status
    virtual_router_name: str
    routes: Dict[str, Route] = field(default_factory=dict)
    tags: List[Dict[str, str]] = field(default_factory=list)

    def to_dict(self) -> Dict[str, Any]:  # type ignore[misc]
        return {
            "meshName": self.mesh_name,
            "metadata": {
                "arn": self.metadata.arn,
                "createdAt": self.metadata.created_at.strftime("%d/%m/%Y, %H:%M:%S"),
                "lastUpdatedAt": self.metadata.last_updated_at.strftime(
                    "%d/%m/%Y, %H:%M:%S"
                ),
                "meshOwner": self.metadata.mesh_owner,
                "resourceOwner": self.metadata.resource_owner,
                "uid": self.metadata.uid,
                "version": self.metadata.version,
            },
            "spec": {
                "listeners": [
                    {
                        "portMapping": {
                            "port": listener["port_mapping"].port,
                            "protocol": listener["port_mapping"].protocol,
                        }
                    }
                    for listener in self.spec.listeners
                ]
            },
            "status": self.status,
            "virtualRouterName": self.virtual_router_name,
        }



"""opensearchserverless base URL and path."""

from .responses import OpenSearchServiceServerlessResponse

url_bases = [
    r"https?://aoss\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": OpenSearchServiceServerlessResponse.dispatch,
}



import json
from copy import deepcopy

import boto3

from moto import mock_aws

template_create_table = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Resources": {
        "table": {
            "Type": "AWS::DynamoDB::Table",
            "Properties": {
                "AttributeDefinitions": [
                    {"AttributeName": "Name", "AttributeType": "S"},
                    {"AttributeName": "Age", "AttributeType": "S"},
                ],
                "KeySchema": [
                    {"AttributeName": "Name", "KeyType": "HASH"},
                    {"AttributeName": "Age", "KeyType": "RANGE"},
                ],
                "ProvisionedThroughput": {
                    "ReadCapacityUnits": 5,
                    "WriteCapacityUnits": 5,
                },
                "TableName": "Person",
            },
        }
    },
}


@mock_aws
def test_create_stack_pay_per_request():
    conn = boto3.client("cloudformation", region_name="us-east-1")
    dynamodb_client = boto3.client("dynamodb", region_name="us-east-1")
    template = deepcopy(template_create_table)
    template["Resources"]["table"]["Properties"]["BillingMode"] = "PAY_PER_REQUEST"
    del template["Resources"]["table"]["Properties"]["ProvisionedThroughput"]

    conn.create_stack(StackName="test", TemplateBody=json.dumps(template))
    table_desc = dynamodb_client.list_tables()
    assert len(table_desc.get("TableNames")) == 1

    table = dynamodb_client.describe_table(TableName=table_desc["TableNames"][0])[
        "Table"
    ]
    assert table["BillingModeSummary"] == {"BillingMode": "PAY_PER_REQUEST"}


@mock_aws
def test_create_stack_with_indexes():
    conn = boto3.client("cloudformation", region_name="us-east-1")
    dynamodb_client = boto3.client("dynamodb", region_name="us-east-1")
    template = deepcopy(template_create_table)
    template["Resources"]["table"]["Properties"]["GlobalSecondaryIndexes"] = [
        {
            "IndexName": "gsi",
            "KeySchema": [{"AttributeName": "gsipk", "KeyType": "S"}],
            "Projection": {"ProjectionType": "ALL"},
        }
    ]
    template["Resources"]["table"]["Properties"]["LocalSecondaryIndexes"] = [
        {
            "IndexName": "lsi",
            "KeySchema": [{"AttributeName": "lsipk", "KeyType": "S"}],
            "Projection": {"ProjectionType": "ALL"},
        }
    ]

    conn.create_stack(StackName="test", TemplateBody=json.dumps(template))
    table_desc = dynamodb_client.list_tables()
    assert len(table_desc.get("TableNames")) == 1

    table = dynamodb_client.describe_table(TableName=table_desc["TableNames"][0])[
        "Table"
    ]
    assert len(table["GlobalSecondaryIndexes"]) == 1
    assert len(table["LocalSecondaryIndexes"]) == 1


@mock_aws
def test_delete_stack_dynamo_template():
    conn = boto3.client("cloudformation", region_name="us-east-1")
    dynamodb_client = boto3.client("dynamodb", region_name="us-east-1")

    conn.create_stack(
        StackName="test_stack", TemplateBody=json.dumps(template_create_table)
    )
    table_desc = dynamodb_client.list_tables()
    assert len(table_desc.get("TableNames")) == 1

    table = dynamodb_client.describe_table(TableName=table_desc["TableNames"][0])[
        "Table"
    ]
    assert table["ProvisionedThroughput"] == {
        "NumberOfDecreasesToday": 0,
        "ReadCapacityUnits": 5,
        "WriteCapacityUnits": 5,
    }
    assert table["BillingModeSummary"] == {"BillingMode": "PROVISIONED"}
    assert table["LocalSecondaryIndexes"] == []
    assert table["GlobalSecondaryIndexes"] == []
    assert table["DeletionProtectionEnabled"] is False

    conn.delete_stack(StackName="test_stack")
    table_desc = dynamodb_client.list_tables()
    assert len(table_desc.get("TableNames")) == 0

    conn.create_stack(
        StackName="test_stack", TemplateBody=json.dumps(template_create_table)
    )



import boto3

from moto import mock_aws


@mock_aws
def test_tag_resource():
    client = boto3.client("inspector2", region_name="ap-southeast-1")
    arn = client.create_filter(
        name="my_first_filter",
        action="NONE",
        filterCriteria={"findingArn": [{"comparison": "EQUALS", "value": "cvdn"}]},
    )["arn"]

    assert client.list_tags_for_resource(resourceArn=arn)["tags"] == {}

    client.tag_resource(resourceArn=arn, tags={"k1": "v1"})
    assert client.list_tags_for_resource(resourceArn=arn)["tags"] == {"k1": "v1"}

    client.tag_resource(resourceArn=arn, tags={"k2": "v2"})
    assert client.list_tags_for_resource(resourceArn=arn)["tags"] == {
        "k1": "v1",
        "k2": "v2",
    }

    client.untag_resource(resourceArn=arn, tagKeys=["k1"])
    assert client.list_tags_for_resource(resourceArn=arn)["tags"] == {"k2": "v2"}


@mock_aws
def test_tag_filter():
    client = boto3.client("inspector2", region_name="ap-southeast-1")
    arn = client.create_filter(
        name="my_first_filter",
        action="NONE",
        filterCriteria={"findingArn": [{"comparison": "EQUALS", "value": "cvdn"}]},
        tags={"k1": "v1"},
    )["arn"]

    assert client.list_tags_for_resource(resourceArn=arn)["tags"] == {"k1": "v1"}

    client.tag_resource(resourceArn=arn, tags={"k2": "v2"})
    assert client.list_tags_for_resource(resourceArn=arn)["tags"] == {
        "k1": "v1",
        "k2": "v2",
    }

    filters = client.list_filters()["filters"]
    assert filters[0]["tags"] == {"k1": "v1", "k2": "v2"}



import io
from unittest.mock import patch
from urllib.parse import parse_qs, urlparse

import pytest
import requests
import xmltodict
from flask.testing import FlaskClient

import moto.server as server
from moto.moto_server.threaded_moto_server import ThreadedMotoServer


class AuthenticatedClient(FlaskClient):
    def open(self, *args, **kwargs):
        kwargs["headers"] = kwargs.get("headers", {})
        kwargs["headers"]["Authorization"] = "Any authorization header"
        kwargs["content_length"] = 0  # Fixes content-length complaints.
        return super().open(*args, **kwargs)


def authenticated_client():
    backend = server.create_backend_app("s3")
    backend.test_client_class = AuthenticatedClient
    return backend.test_client()


def test_s3_server_get():
    test_client = authenticated_client()
    res = test_client.get("/")

    assert b"ListAllMyBucketsResult" in res.data


@pytest.mark.parametrize("key_name", ["bar_baz", "bar+baz", "baz bar"])
def test_s3_server_bucket_create(key_name):
    test_client = authenticated_client()

    res = test_client.put("/", "http://foobaz.localhost:5000/")
    assert res.status_code == 200

    res = test_client.get("/")
    assert b"<Name>foobaz</Name>" in res.data

    res = test_client.get("/", "http://foobaz.localhost:5000/")
    assert res.status_code == 200
    assert b"ListBucketResult" in res.data

    res = test_client.put(
        f"/{key_name}", "http://foobaz.localhost:5000/", data="test value"
    )
    assert res.status_code == 200
    assert "ETag" in dict(res.headers)

    # ListObjects
    res = test_client.get(
        "/", "http://foobaz.localhost:5000/", query_string={"prefix": key_name}
    )
    assert res.status_code == 200
    content = xmltodict.parse(res.data)["ListBucketResult"]["Contents"]
    # If we receive a dict, we only received one result
    # If content is of type list, our call returned multiple results - which is not correct
    assert isinstance(content, dict)
    assert content["Key"] == key_name

    # ListObjects V2
    res = test_client.get(
        "/",
        "http://foobaz.localhost:5000/",
        query_string={"prefix": key_name, "list-type": "2"},
    )
    assert res.status_code == 200
    content = xmltodict.parse(res.data)["ListBucketResult"]["Contents"]
    assert isinstance(content, dict)
    assert content["Key"] == key_name

    # ListObjectsVersions
    res = test_client.get(
        "/",
        "http://foobaz.localhost:5000/",
        query_string={"prefix": key_name, "versions": ""},
    )
    assert res.status_code == 200
    assert xmltodict.parse(res.data)["ListVersionsResult"]["Prefix"] == key_name
    content = xmltodict.parse(res.data)["ListVersionsResult"]["Version"]
    assert isinstance(content, dict)
    assert content["Key"] == key_name

    # GetBucket
    res = test_client.head("http://foobaz.localhost:5000")
    assert res.status_code == 200
    assert res.headers.get("x-amz-bucket-region") == "us-east-1"

    # HeadObject
    res = test_client.head(f"/{key_name}", "http://foobaz.localhost:5000/")
    assert res.status_code == 200
    assert res.headers.get("Accept-Ranges") == "bytes"

    # GetObject
    res = test_client.get(f"/{key_name}", "http://foobaz.localhost:5000/")
    assert res.status_code == 200
    assert res.data == b"test value"
    assert res.headers.get("Accept-Ranges") == "bytes"


def test_s3_server_ignore_subdomain_for_bucketnames():
    with patch("moto.settings.S3_IGNORE_SUBDOMAIN_BUCKETNAME", True):
        test_client = authenticated_client()

        res = test_client.put("/mybucket", "http://foobaz.localhost:5000/")
        assert res.status_code == 200
        assert b"mybucket" in res.data


def test_s3_server_bucket_versioning():
    test_client = authenticated_client()

    res = test_client.put("/", "http://foobaz.localhost:5000/")
    assert res.status_code == 200

    # Just enough XML to enable versioning
    body = "<Status>Enabled</Status>"
    res = test_client.put("/?versioning", "http://foobaz.localhost:5000", data=body)
    assert res.status_code == 200


def test_s3_server_post_to_bucket():
    test_client = authenticated_client()

    res = test_client.put("/", "http://tester.localhost:5000/")
    assert res.status_code == 200

    test_client.post(
        "/",
        "https://tester.localhost:5000/",
        data={"key": "the-key", "file": "nothing"},
    )

    res = test_client.get("/the-key", "http://tester.localhost:5000/")
    assert res.status_code == 200
    assert res.data == b"nothing"


def test_s3_server_post_to_bucket_redirect():
    test_client = authenticated_client()

    res = test_client.put("/", "http://tester.localhost:5000/")
    assert res.status_code == 200

    redirect_base = "https://redirect.com/success/"
    filecontent = "nothing"
    filename = "test_filename.txt"
    res = test_client.post(
        "/",
        "https://tester.localhost:5000/",
        data={
            "key": "asdf/the-key/${filename}",
            "file": (io.BytesIO(filecontent.encode("utf8")), filename),
            "success_action_redirect": redirect_base,
        },
    )
    real_key = f"asdf/the-key/{filename}"
    assert res.status_code == 303
    redirect = res.headers["location"]
    assert redirect.startswith(redirect_base)

    parts = urlparse(redirect)
    args = parse_qs(parts.query)
    assert args["key"][0] == real_key
    assert args["bucket"][0] == "tester"

    res = test_client.get(f"/{real_key}", "http://tester.localhost:5000/")
    assert res.status_code == 200
    assert res.data == filecontent.encode("utf8")


def test_s3_server_post_without_content_length():
    test_client = authenticated_client()

    # You can create a bucket without specifying Content-Length
    res = test_client.put(
        "/", "http://tester.localhost:5000/", environ_overrides={"CONTENT_LENGTH": ""}
    )
    assert res.status_code == 200

    # You can specify a bucket in another region without specifying Content-Length
    # (The body is just ignored..)
    res = test_client.put(
        "/",
        "http://tester.localhost:5000/",
        environ_overrides={"CONTENT_LENGTH": ""},
        data=(
            "<CreateBucketConfiguration>"
            "<LocationConstraint>us-west-2</LocationConstraint>"
            "</CreateBucketConfiguration>"
        ),
    )
    assert res.status_code == 200

    # You cannot make any other bucket-related requests without specifying Content-Length
    for path in ["/?versioning", "/?policy"]:
        res = test_client.put(
            path, "http://t.localhost:5000", environ_overrides={"CONTENT_LENGTH": ""}
        )
        assert res.status_code == 411

    # You cannot make any POST-request
    res = test_client.post(
        "/", "https://tester.localhost:5000/", environ_overrides={"CONTENT_LENGTH": ""}
    )
    assert res.status_code == 411


def test_s3_server_post_unicode_bucket_key():
    """Verify non-ascii characters in request URLs (e.g., S3 object names)."""
    dispatcher = server.DomainDispatcherApplication(server.create_backend_app)
    backend_app = dispatcher.get_application(
        {
            "HTTP_HOST": "s3.amazonaws.com",
            "PATH_INFO": "/test-bucket/test-object-てすと",
        }
    )
    assert backend_app
    backend_app = dispatcher.get_application(
        {
            "HTTP_HOST": "s3.amazonaws.com",
            "PATH_INFO": "/test-bucket/test-object-てすと".encode("utf-8"),
        }
    )
    assert backend_app


def test_s3_server_post_cors():
    """Test default CORS headers set by flask-cors plugin"""
    test_client = authenticated_client()
    # Create the bucket
    test_client.put("/", "http://tester.localhost:5000/")

    preflight_headers = {
        "Access-Control-Request-Method": "POST",
        "Access-Control-Request-Headers": "origin, x-requested-with",
        "Origin": "https://localhost:9000",
    }

    res = test_client.options(
        "/", "http://tester.localhost:5000/", headers=preflight_headers
    )
    assert res.status_code in [200, 204]

    expected_methods = set(["DELETE", "PATCH", "PUT", "GET", "HEAD", "POST", "OPTIONS"])
    assert (
        set(res.headers["Access-Control-Allow-Methods"].split(", ")) == expected_methods
    )

    assert res.headers["Access-Control-Allow-Origin"] == "https://localhost:9000"
    assert res.headers["Access-Control-Allow-Headers"] == "origin, x-requested-with"


def test_s3_server_post_cors_exposed_header():
    """Test overriding default CORS headers with custom bucket rules"""
    # github.com/getmoto/moto/issues/4220

    cors_config_payload = """<CORSConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
  <CORSRule>
    <AllowedOrigin>https://example.org</AllowedOrigin>
    <AllowedMethod>HEAD</AllowedMethod>
    <AllowedMethod>GET</AllowedMethod>
    <AllowedMethod>PUT</AllowedMethod>
    <AllowedMethod>POST</AllowedMethod>
    <AllowedMethod>DELETE</AllowedMethod>
    <AllowedHeader>*</AllowedHeader>
    <ExposeHeader>ETag</ExposeHeader>
    <MaxAgeSeconds>3000</MaxAgeSeconds>
  </CORSRule>
</CORSConfiguration>
    """

    test_client = authenticated_client()
    valid_origin = "https://example.org"
    preflight_headers = {
        "Access-Control-Request-Method": "POST",
        "Access-Control-Request-Headers": "origin, x-requested-with",
        "Origin": valid_origin,
    }
    # Returns 403 on non existing bucket
    preflight_response = test_client.options(
        "/", "http://testcors.localhost:5000/", headers=preflight_headers
    )
    assert preflight_response.status_code == 403

    # Create the bucket & file
    test_client.put("/", "http://testcors.localhost:5000/")
    test_client.put("/test", "http://testcors.localhost:5000/")
    res = test_client.put(
        "/?cors", "http://testcors.localhost:5000", data=cors_config_payload
    )
    assert res.status_code == 200

    cors_res = test_client.get("/?cors", "http://testcors.localhost:5000")
    assert b"<ExposedHeader>ETag</ExposedHeader>" in cors_res.data

    # Test OPTIONS bucket response and key response
    for key_name in ("/", "/test"):
        preflight_response = test_client.options(
            key_name, "http://testcors.localhost:5000/", headers=preflight_headers
        )
        assert preflight_response.status_code == 200
        expected_cors_headers = {
            "Access-Control-Allow-Methods": "HEAD, GET, PUT, POST, DELETE",
            "Access-Control-Allow-Origin": "https://example.org",
            "Access-Control-Allow-Headers": "*",
            "Access-Control-Expose-Headers": "ETag",
            "Access-Control-Max-Age": "3000",
        }
        for header_name, header_value in expected_cors_headers.items():
            assert header_name in preflight_response.headers
            assert preflight_response.headers[header_name] == header_value

    # Test GET key response
    # A regular GET should not receive any CORS headers
    resp = test_client.get("/test", "http://testcors.localhost:5000/")
    assert "Access-Control-Allow-Methods" not in resp.headers
    assert "Access-Control-Expose-Headers" not in resp.headers

    # A GET with mismatched Origin-header should not receive any CORS headers
    resp = test_client.get(
        "/test", "http://testcors.localhost:5000/", headers={"Origin": "something.com"}
    )
    assert "Access-Control-Allow-Methods" not in resp.headers
    assert "Access-Control-Expose-Headers" not in resp.headers

    # Only a GET with matching Origin-header should receive CORS headers
    resp = test_client.get(
        "/test", "http://testcors.localhost:5000/", headers={"Origin": valid_origin}
    )
    assert (
        resp.headers["Access-Control-Allow-Methods"] == "HEAD, GET, PUT, POST, DELETE"
    )
    assert resp.headers["Access-Control-Expose-Headers"] == "ETag"

    # Test PUT key response
    # A regular PUT should not receive any CORS headers
    resp = test_client.put("/test", "http://testcors.localhost:5000/")
    assert "Access-Control-Allow-Methods" not in resp.headers
    assert "Access-Control-Expose-Headers" not in resp.headers

    # A PUT with mismatched Origin-header should not receive any CORS headers
    resp = test_client.put(
        "/test", "http://testcors.localhost:5000/", headers={"Origin": "something.com"}
    )
    assert "Access-Control-Allow-Methods" not in resp.headers
    assert "Access-Control-Expose-Headers" not in resp.headers

    # Only a PUT with matching Origin-header should receive CORS headers
    resp = test_client.put(
        "/test", "http://testcors.localhost:5000/", headers={"Origin": valid_origin}
    )
    assert (
        resp.headers["Access-Control-Allow-Methods"] == "HEAD, GET, PUT, POST, DELETE"
    )
    assert resp.headers["Access-Control-Expose-Headers"] == "ETag"


def test_s3_server_post_cors_multiple_origins():
    """Test that Moto only responds with the Origin that we that hosts the server"""
    # github.com/getmoto/moto/issues/6003

    cors_config_payload = """<CORSConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
  <CORSRule>
    <AllowedOrigin>https://example.org</AllowedOrigin>
    <AllowedOrigin>https://localhost:6789</AllowedOrigin>
    <AllowedMethod>POST</AllowedMethod>
  </CORSRule>
</CORSConfiguration>
    """

    thread = ThreadedMotoServer(port="6789", verbose=False)
    thread.start()

    # Create the bucket
    requests.put("http://testcors.localhost:6789/")
    requests.put("http://testcors.localhost:6789/?cors", data=cors_config_payload)

    # Test only our requested origin is returned
    preflight_response = requests.options(
        "http://testcors.localhost:6789/test2",
        headers={
            "Access-Control-Request-Method": "POST",
            "Origin": "https://localhost:6789",
        },
    )
    assert preflight_response.status_code == 200
    assert (
        preflight_response.headers["Access-Control-Allow-Origin"]
        == "https://localhost:6789"
    )
    assert preflight_response.content == b""

    # Verify a request with unknown origin fails
    preflight_response = requests.options(
        "http://testcors.localhost:6789/test2",
        headers={
            "Access-Control-Request-Method": "POST",
            "Origin": "https://unknown.host",
        },
    )
    assert preflight_response.status_code == 403
    assert b"<Code>AccessForbidden</Code>" in preflight_response.content

    # Verify we can use a wildcard anywhere in the origin
    cors_config_payload = (
        '<CORSConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><CORSRule>'
        "<AllowedOrigin>https://*.google.com</AllowedOrigin>"
        "<AllowedMethod>POST</AllowedMethod>"
        "</CORSRule></CORSConfiguration>"
    )
    requests.put("http://testcors.localhost:6789/?cors", data=cors_config_payload)
    for origin in ["https://sth.google.com", "https://a.google.com"]:
        preflight_response = requests.options(
            "http://testcors.localhost:6789/test2",
            headers={"Access-Control-Request-Method": "POST", "Origin": origin},
        )
        assert preflight_response.status_code == 200
        assert preflight_response.headers["Access-Control-Allow-Origin"] == origin

    # Non-matching requests throw an error though - it does not act as a full wildcard
    preflight_response = requests.options(
        "http://testcors.localhost:6789/test2",
        headers={
            "Access-Control-Request-Method": "POST",
            "Origin": "sth.microsoft.com",
        },
    )
    assert preflight_response.status_code == 403
    assert b"<Code>AccessForbidden</Code>" in preflight_response.content

    # Verify we can use a wildcard as the origin
    cors_config_payload = (
        '<CORSConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><CORSRule>'
        "<AllowedOrigin>*</AllowedOrigin>"
        "<AllowedMethod>POST</AllowedMethod>"
        "</CORSRule></CORSConfiguration>"
    )
    requests.put("http://testcors.localhost:6789/?cors", data=cors_config_payload)
    for origin in ["https://a.google.com", "http://b.microsoft.com", "any"]:
        preflight_response = requests.options(
            "http://testcors.localhost:6789/test2",
            headers={"Access-Control-Request-Method": "POST", "Origin": origin},
        )
        assert preflight_response.status_code == 200
        assert preflight_response.headers["Access-Control-Allow-Origin"] == origin

    thread.stop()



import json
import typing
from datetime import datetime
from typing import Any, DefaultDict, Dict, List, Optional

from dateutil.tz import tzutc

from moto.s3.models import s3_backends
from moto.utilities.utils import get_partition

from .exceptions import ValidationError

if typing.TYPE_CHECKING:
    from .models import FakeModelCard, FakePipeline, FakePipelineExecution


def get_pipeline_from_name(
    pipelines: Dict[str, "FakePipeline"], pipeline_name: str
) -> "FakePipeline":
    try:
        return pipelines[pipeline_name]
    except KeyError:
        raise ValidationError(
            message=f"Could not find pipeline with PipelineName {pipeline_name}."
        )


def get_pipeline_name_from_execution_arn(pipeline_execution_arn: str) -> str:
    return pipeline_execution_arn.split("/")[1].split(":")[-1]


def get_pipeline_execution_from_arn(
    pipelines: Dict[str, "FakePipeline"], pipeline_execution_arn: str
) -> "FakePipelineExecution":
    try:
        pipeline_name = get_pipeline_name_from_execution_arn(pipeline_execution_arn)
        pipeline = get_pipeline_from_name(pipelines, pipeline_name)
        return pipeline.pipeline_executions[pipeline_execution_arn]
    except KeyError:
        raise ValidationError(
            message=f"Could not find pipeline execution with PipelineExecutionArn {pipeline_execution_arn}."
        )


def load_pipeline_definition_from_s3(
    pipeline_definition_s3_location: Dict[str, Any], account_id: str, partition: str
) -> Dict[str, Any]:
    s3_backend = s3_backends[account_id][partition]
    result = s3_backend.get_object(
        bucket_name=pipeline_definition_s3_location["Bucket"],
        key_name=pipeline_definition_s3_location["ObjectKey"],
    )
    return json.loads(result.value)  # type: ignore[union-attr]


def arn_formatter(_type: str, _id: str, account_id: str, region_name: str) -> str:
    return f"arn:{get_partition(region_name)}:sagemaker:{region_name}:{account_id}:{_type}/{_id}"


def validate_model_approval_status(model_approval_status: typing.Optional[str]) -> None:
    if model_approval_status is not None and model_approval_status not in [
        "Approved",
        "Rejected",
        "PendingManualApproval",
    ]:
        raise ValidationError(
            f"Value '{model_approval_status}' at 'modelApprovalStatus' failed to satisfy constraint: "
            "Member must satisfy enum value set: [PendingManualApproval, Approved, Rejected]"
        )


def filter_model_cards(
    model_cards: DefaultDict[str, List["FakeModelCard"]],
    creation_time_after: Optional[datetime],
    creation_time_before: Optional[datetime],
    name_contains: Optional[str],
    model_card_status: Optional[str],
    sort_by: Optional[str],
    sort_order: Optional[str],
) -> List["FakeModelCard"]:
    reverse = sort_order == "Descending"

    if name_contains:
        lowercase_name = name_contains.lower()
        filtered_cards = {
            k: v for k, v in model_cards.items() if lowercase_name in k.lower()
        }
    else:
        filtered_cards = {k: v for k, v in model_cards.items()}

    result: List[FakeModelCard] = []
    for _, versions in filtered_cards.items():
        filtered_versions = versions

        if creation_time_after:
            if isinstance(creation_time_after, int):
                creation_time_after = datetime.fromtimestamp(
                    creation_time_after, tz=tzutc()
                )
            filtered_versions = [
                v
                for v in filtered_versions
                if v.last_modified_time > str(creation_time_after)
            ]

        if creation_time_before:
            if isinstance(creation_time_before, int):
                creation_time_before = datetime.fromtimestamp(
                    creation_time_before, tz=tzutc()
                )
            filtered_versions = [
                v
                for v in filtered_versions
                if v.last_modified_time < str(creation_time_before)
            ]

        if model_card_status:
            filtered_versions = [
                v for v in filtered_versions if v.model_card_status == model_card_status
            ]

        if filtered_versions:
            latest_version = max(filtered_versions, key=lambda v: v.last_modified_time)
            result.append(latest_version)

    if not result:
        return []

    def sort_key(x: "FakeModelCard") -> str:
        if sort_by == "Name":
            return x.model_card_name
        return x.creation_time

    return sorted(result, key=sort_key, reverse=reverse)
